exception Unitialised_variable of string ;;
exception Undefined_variable of string ;;
exception Variable_not_input of string ;;
exception Type_mismatch of string ;;




(* ************************************************************************* *)
(*  [Fonc] dump_variable_declaration :                                       *)
(*    out_channel: out_channel -> trailer:string ->                          *)
(*      check_mutability:bool -> declaration:Basics.variable_description ->  *)
(*	  apply_extern:bool -> unit                                          *)
(** [Description] :

    [Args] :
      - trailer : Chane de caractres  mettre en fin de dclaration.
      - check_mutability : Spcifie si l'on doit tweaker la dclaration en
          lui rajoutant une "*" si elle est OUT ou INOUT. Pragmatiquement,
          c'est le cas lorsque l'on gnre une dclaration de paramtres et
          ce n'est pas le cas pour les autres dclarations (variables
          globales et variables locales du "main").
          Notons que ceci ne s'applique pas aux variables de type chanes,
          celles-ci tant dj des pointeurs en C.
      - apply_extern : Spcifie si l'on doit mentionner la dclaration comme
          "extern" ou non. Pragmatiquement, c'est le cas lorsque l'on gnre
          un proto de variable globale. Sinon, ce n'est pas le cas.          *)
(* ************************************************************************* *)
let dump_variable_declaration ~out_channel ~trailer
    ~check_mutability ~declaration: var_descr ~apply_extern =
  let string_of_mut =
    (match (check_mutability, var_descr.Basics.var_mutability) with
     | (false, _) | (true, Basics.IN) -> ""
     | (true, Basics.OUT) | (true, Basics.INOUT) ->
	 begin
	 (* Une variable en mode OUT ou INOUT est    *)
	 (* implicitement modlise par un pointeur. *)
	 (match var_descr.Basics.var_type with
	 | Basics.INT -> "*"
	 | Basics.INT_ARRAY _ -> ""
	 | Basics.STRING  -> ""
	 | Basics.TUPLE  -> "*" )
	 end) in
  let s_extern = if apply_extern then "extern " else ""  in
  match var_descr.Basics.var_type with
   | Basics.INT ->
       Printf.fprintf
	 out_channel "%sint %s%s%s"
	 s_extern string_of_mut var_descr.Basics.var_name trailer
   | Basics.INT_ARRAY size ->
       Printf.fprintf
	 out_channel "%sint %s%s[%d]%s"
	 s_extern string_of_mut var_descr.Basics.var_name size trailer
   | Basics.STRING ->
       Printf.fprintf
	 out_channel "%schar %s[42]%s"
	 s_extern var_descr.Basics.var_name trailer
   | Basics.TUPLE ->
       Printf.fprintf
	 out_channel "%stuple %s%s%s" 
	 s_extern string_of_mut var_descr.Basics.var_name trailer

;;





(* ********************************************************************* *)
(*  [Fonc] dump_parameter_declarations :                                 *)
(*           out_channel:out_channel -> trailer:string ->                *)
(*	       declarations:Basics.variable_description list -> unit     *)
(** [Description] :

    [Args] :
      - trailer : Chane de caractres  mettre en fin de chaque
           dclaration.                                                  *)
(* ********************************************************************* *)
let rec dump_parameter_declarations ~out_channel ~trailer
    ~declarations: decls =
  match decls with
  | [] -> ()
  | [last] ->
      dump_variable_declaration
	~out_channel ~trailer: "" ~check_mutability: true
	~declaration: last ~apply_extern: false
  | h :: q ->
      dump_variable_declaration
	~out_channel ~trailer
	~check_mutability: true ~declaration: h ~apply_extern: false ;
      dump_parameter_declarations
	~out_channel ~trailer ~declarations: q
;;




(* trailer : Chane de caractres  mettre en fin d'argument. *)
let dump_arg ~out_channel ~trailer ~arg =
  let string_of_mut =
    (match arg.Basics.var_mutability with
     | Basics.IN -> ""
     | Basics.OUT
     | Basics.INOUT -> 
	 (match arg.Basics.var_type with
	 | Basics.INT -> "&"
	 | Basics.INT_ARRAY _ -> ""
	 | Basics.STRING -> ""
	 | Basics.TUPLE -> "&" )
    )in
  Printf.fprintf
    out_channel "%s%s%s" string_of_mut arg.Basics.var_name trailer
;;



(* ********************************************************************** *)
(* [Fonc] dump_call_args :                                                *)
(*          out_channel:out_channel ->                                    *)
(*            args:Basics.variable_description list -> unit               *)
(** [Description] : Dumpe une liste de variable comme tant des arguments
        d'appel. Chaque variable est donc spare de sa suivante par une
        virgule, sauf la dernire, videmment !                           *)
(* ********************************************************************** *)
let rec dump_call_args ~out_channel ~args =
  match args with
  | [] -> ()
  | [last] ->
      dump_arg ~out_channel ~trailer: "" ~arg: last
  | h :: q ->
      dump_arg ~out_channel ~trailer: ", " ~arg: h ;
      dump_call_args ~out_channel ~args: q
;;




(* seen : liste des variables ayant dj t rinitialises. On ne rinitialise pas plusieurs fois la mme variables dans un mme scnario. *)
let reset_output ~out_channel ~seen ~variable =
  let init_val =
    (match variable.Basics.var_init with
     | None -> raise (Unitialised_variable variable.Basics.var_name)
     | Some i -> i) in
  if List.mem variable.Basics.var_name seen then seen
  else
    begin
    (match variable.Basics.var_type with
     | Basics.TUPLE ->
	 (begin
	   (match init_val with
	   | Basics.Init_integer _ ->
	       raise (Type_mismatch variable.Basics.var_name)
	   | Basics.Init_string _ ->
	       raise (Type_mismatch variable.Basics.var_name)
	   | Basics.Init_tuple (ch1,ch2) -> 
	       Printf.fprintf out_channel "%s.first = %s ;\n"
		 variable.Basics.var_name (Int32.to_string ch1);
	       Printf.fprintf out_channel "%s.second = %s ;\n"
		 variable.Basics.var_name (Int32.to_string ch2) ) 
	 end)
     | Basics.STRING ->
	 (begin
	   let string_val =
	     (match init_val with
	     | Basics.Init_integer _ ->
		 raise (Type_mismatch variable.Basics.var_name)
	     | Basics.Init_string i -> i
	     | Basics.Init_tuple (_,_) ->
		 raise (Type_mismatch variable.Basics.var_name) ) 
	   in 
	   Printf.fprintf out_channel "strcpy (%s, %s) ;\n"
	     variable.Basics.var_name string_val
	 end)
     | Basics.INT ->
	 (begin
	   let integer_val =
	     (match init_val with
	     | Basics.Init_integer i -> i
	     | Basics.Init_string _ ->
		 raise (Type_mismatch variable.Basics.var_name)
	     | Basics.Init_tuple _ ->
		 raise (Type_mismatch variable.Basics.var_name) ) 
	   in
	   Printf.fprintf out_channel "%s = %s ;\n"
	     variable.Basics.var_name (Int32.to_string integer_val)
	 end)
     | Basics.INT_ARRAY size ->
	 (begin
	   let integer_val =
	     (match init_val with
	     | Basics.Init_integer i -> i
	     | Basics.Init_string _ ->
		 raise (Type_mismatch variable.Basics.var_name)
 	     | Basics.Init_tuple _ ->
		 raise (Type_mismatch variable.Basics.var_name) ) 
	   in
	   Printf.fprintf
	     out_channel "memset (%s, %s, %d) ;\n"
	     variable.Basics.var_name (Int32.to_string integer_val) size
	 end)) ;
	 variable.Basics.var_name :: seen
	 end
;;





let set_input ~out_channel: out_ch ~shot ~seen ~input: (var, value) =
  (* On recherche la variable parmi celles dclares. *)
  let var_name =
    (match var with 
    | Basics.Var_atomic s -> s 
    | Basics.Var_array (s, _) -> s) in
  let vd =
    try List.find
	(fun v -> v.Basics.var_name = var_name)
	(shot.Basics.tested_parameters @ shot.Basics.tested_globvars) with
    | Not_found -> raise (Undefined_variable var_name) in
  (* On s'assure qu'elle est bien en IN ou en INOUT. *)
  if vd.Basics.var_mutability = Basics.OUT then
    raise (Variable_not_input var_name) ;
  (* Maintenant, on regarde s'il y a besoin de faire une init globale *)
  (* (effacement complet d'un tableau avant de pouvoir initialiser    *)
  (* ensuite chaque case explicitement mentionne.                    *)
  let new_seen =
    if List.mem vd.Basics.var_name seen then seen
    else
      begin
      (match vd.Basics.var_type with
       | Basics.INT -> ()
       | Basics.INT_ARRAY size ->
	   (begin
	   match vd.Basics.var_init with
	    | None -> raise (Unitialised_variable var_name)
	    | Some (Basics.Init_integer i) ->
		Printf.fprintf
		  out_ch "memset (%s, %s, %d) ;\n"
		  vd.Basics.var_name (Int32.to_string i) size
	    | Some (Basics.Init_string _) ->
		(* On ne gre pas les tableaux de chanes. *)
		raise (Type_mismatch vd.Basics.var_name)
	    | Some (Basics.Init_tuple _) ->
		(* On ne gre pas les tableaux de couples. *)
		raise (Type_mismatch vd.Basics.var_name)
           end)
       | Basics.STRING ->
	   (begin
	   match vd.Basics.var_init with
	    | None -> raise (Unitialised_variable vd.Basics.var_name)
	    | Some (Basics.Init_string s) ->
		Printf.fprintf
		  out_ch "strcpy (%s, %s) ;\n" vd.Basics.var_name s
	    | Some (Basics.Init_integer _) ->
		raise (Type_mismatch vd.Basics.var_name)
	    | Some (Basics.Init_tuple _) ->
		raise (Type_mismatch vd.Basics.var_name)
	   end)
       | Basics.TUPLE -> ()) ;
	vd.Basics.var_name :: seen
      end in
  (* Maintenant on affecte la valeur spcifice dans le scnario de test. *)
  (match var with
   | Basics.Var_atomic s ->
       (begin
       match value with
	| Basics.Init_integer i ->
	    Printf.fprintf out_ch "%s = %s ;\n" s (Int32.to_string i)
	| Basics.Init_string str_literal ->
	    Printf.fprintf out_ch "strcpy (%s, %s) ;\n" s str_literal
 	| Basics.Init_tuple (ch1, ch2) ->
	    Printf.fprintf out_ch "%s.first = %s ;\n" s (Int32.to_string ch1);
	    Printf.fprintf out_ch "%s.second = %s ;\n" s (Int32.to_string ch2)
       end)
   | Basics.Var_array (s, index) ->
       (begin
       match value with
        | Basics.Init_integer i ->
	    Printf.fprintf out_ch "%s[%d] = %s ;\n" s index (Int32.to_string i)
	| Basics.Init_string _ ->
	    (* On ne gre pas les tableaux de chanes. *)
	    raise (Type_mismatch vd.Basics.var_name)
	| Basics.Init_tuple _ ->
	    (* On ne gre pas les tableaux de couples. *)
	    raise (Type_mismatch vd.Basics.var_name)
       end)) ;
  (* Et on renvoie le nouvel ensemble de variables vues et donc *)
  (* si ncessaires dj globalement affectes par un memset.   *)
  new_seen
;;




let dump_scenario_IO_summary ~out_channel ~scenario =
  (* Header dcrivant le test et le mot "Entres". *)
  Printf.fprintf out_channel "\n%s\n"
    (Messages.msg_Test_header
       scenario.Basics.test_number
       (String.escaped scenario.Basics.test_comment)) ;
  (* Gnration du code d'affichage des entres. *)
  List.iter
    (fun (var, value) ->
      let string_of_value =
	(match value with
	 | Basics.Init_string s ->
	     (* Attention, la chane 's' contient dj ses guillemets. *)
	     (* Donc celui qui la termine va se retrouver directement  *)
	     (* inlin dans le printf et va pter le format. On va     *)
	     (* donc insrer un backslash devant avant de la filer    *)
	     (* printf.                                                *)
	     let s_without_trailing_quote =
	       String.sub s 0 ((String.length s) - 1) in
	     "\\" ^ s_without_trailing_quote ^ "\\\""
	 | Basics.Init_integer i -> Int32.to_string i
	 | Basics.Init_tuple (ch1, ch2) ->
 	     ("(" ^ (Int32.to_string ch1) ^ ", " ^
	      (Int32.to_string ch2)^")")) in
      match var with
       | Basics.Var_atomic s ->
	   Printf.fprintf out_channel "printf (\"  %s : %s\\n\") ;\n"
	     s string_of_value
       | Basics.Var_array (s, index) ->
           Printf.fprintf out_channel
	     "printf (\"  %s[%d] : %s\\n\") ;\n" s index string_of_value)
    scenario.Basics.test_inputs ;
  (* Gnration du code d'affichage des sorties. *)
  Printf.fprintf out_channel "\n%s\n" Messages.msg_Outputs ;
  List.iter
    (fun (var, value) ->
      match var with
       | Basics.Var_atomic s ->
	   (begin
	   match value with
	    | Basics.Init_integer i ->
		(* Attention, il faut afficher un entier -> %d ! *)
		Printf.fprintf out_channel
		  "printf (\"  %s : %s -> %%d\\n\", %s) ;\n"
		  s (Int32.to_string i) s
	    | Basics.Init_string str_literal ->
		(* Attention, il faut afficher une chane -> %s ! *)
		Printf.fprintf out_channel
		  "printf (\"  %s : %s -> \\\"%%s\\\"\\n\", %s) ;\n"
		  s (String.escaped str_literal) s
	    | Basics.Init_tuple (ch1, ch2) ->
		(* Attention, il faut afficher un couple -> (%d, %d) ! *)
		Printf.fprintf out_channel
		  "printf (\"  %s : (%s, %s) -> ( %%d, %%d) \\n\", 
		  %s.first, %s.second ) ;\n"
		  s (Int32.to_string ch1) (Int32.to_string ch2) s s
	   end)
       | Basics.Var_array (s, index) ->
	   (begin
	   match value with
	    | Basics.Init_integer i ->
		Printf.fprintf out_channel
		  "printf (\"  %s[%d] : %s -> %%d\\n\", %s[%d]) ;\n"
		  s index (Int32.to_string i) s index
	    | Basics.Init_string _ ->
		(* On ne gre pas les tableaux de chanes. *)
		raise (Type_mismatch s)
	    | Basics.Init_tuple _ ->
		(* On ne gre pas les tableaux de couples. *)
		raise (Type_mismatch s)
	   end))
    scenario.Basics.test_results
;;





let dump_ands_list ~out_channel: out_ch ~expected =
  let rec dump = function
    | [] -> ()
    | [(var, value)] ->
	(begin
	match var with
	 | Basics.Var_atomic s ->
	     (begin
	     match value with
	      | Basics.Init_integer i -> 
		  Printf.fprintf out_ch "(%s == %s)" s (Int32.to_string i)
	      | Basics.Init_string str_literal ->
		  (* Attention, on ne compare pas des chanes avec ==. *)
		  Printf.fprintf out_ch "(strcmp (%s, %s) == 0)" s str_literal
		  (* Attention, on ne compare pas des couples  ==. *)
	      | Basics.Init_tuple (ch1, ch2) -> 
		  Printf.fprintf out_ch "(%s.first == %s) && " s 
		    (Int32.to_string ch1) ;
		  Printf.fprintf out_ch "(%s.second == %s)" s 
		    (Int32.to_string ch2)
	       end)
	 | Basics.Var_array (s, index) ->
	     (begin
	     let string_of_value =
	       (match value with
		| Basics.Init_integer i -> Int32.to_string i
		| Basics.Init_string _ ->
		    (* On ne gre pas les tableaux de chanes. *)
		    raise (Type_mismatch s)
		| Basics.Init_tuple _ ->
		    (* On ne gre pas les tableaux de couples. *)
		    raise (Type_mismatch s)) in
             Printf.fprintf out_ch "(%s[%d] == %s)" s index string_of_value
	     end)
	end)
    | (var, value) :: q ->
	(begin
	(match var with
	 | Basics.Var_atomic s ->
	     (begin
	     match value with
	      | Basics.Init_integer i -> 
		  Printf.fprintf out_ch "(%s == %s) && " s (Int32.to_string i)
	      | Basics.Init_string str_literal ->
		  (* Attention, on ne compare pas des chanes avec ==. *)
		  Printf.fprintf out_ch "(strcmp (%s, %s) == 0) && "
		    s str_literal
		  (* Attention, on ne compare pas des couples  ==. *)
	      | Basics.Init_tuple (ch1, ch2) -> 
		  Printf.fprintf out_ch "(%s.first == %s) && " s 
		    (Int32.to_string ch1);
		  Printf.fprintf out_ch "(%s.second == %s) && " s 
		    (Int32.to_string ch2)
	       end)
	 | Basics.Var_array (s, index) ->
	     let string_of_value =
	       (match value with
		| Basics.Init_integer i -> Int32.to_string i
		| Basics.Init_string _ ->
		    (* On ne gre pas les tableaux de chanes. *)
		    raise (Type_mismatch s)
		| Basics.Init_tuple _ ->
		    (* On ne gre pas les tableaux de couples. *)
		    raise (Type_mismatch s)) in
             Printf.fprintf out_ch "(%s[%d] == %s) && "
	       s index string_of_value) ;
	(* On dumpe le reste de la condition. *)
	dump q
	end) in
  dump expected
;;





(* ******************************************************************** *)
(*  [Fonc] dump_outputs_check : out_channel:out_channel ->              *)
(*            scenario:Basics.test -> unit                              *)
(** [Description] : Gnration du code effectuant les vrification pour
        un scnario de l'adquation entre sorties obtenues et sorties
        attendues.                                                      *)
(* ******************************************************************** *)
let dump_outputs_check ~out_channel ~scenario =
  Printf.fprintf out_channel "if (" ;
  dump_ands_list ~out_channel ~expected: scenario.Basics.test_results ;
  Printf.fprintf out_channel ") {
  printf (\"*** test OK ***\\n\") ;
  nb_tests++;
  nb_tests_OK++;
  }
else {
  printf (\"!!! test KO !!!\\n\") ;
  nb_tests++;
  nb_tests_KO++;
  }\n ";
  Printf.fprintf out_channel "%s\n" Messages.msg_Test_separator
;;





let dump_tests ~out_channel ~shot =
  let rec dump_test scenario =
    (* Numro du jeu de test. *)
    Printf.fprintf
      out_channel "%s\n"
      (Messages.msg_Test_number scenario.Basics.test_number) ;
    (* Effacement des sorties: variables globales *)
    (* et paramtres en OUT et INOUT.             *)
    let already_reset_output_names = ref [] in
    List.iter
      (fun vd ->
	match vd.Basics.var_mutability with
	 | Basics.OUT | Basics.INOUT ->
	     already_reset_output_names :=
	       reset_output
		 ~out_channel ~seen: !already_reset_output_names
		 ~variable: vd
	 | Basics.IN -> ())
      (shot.Basics.tested_globvars @ shot.Basics.tested_parameters) ;
    (* Positionnement des entres du test: vars globales et params E et ES. *)
    Printf.fprintf out_channel "\n%s\n" Messages.comment_Inputs_settings ;
    let already_set_input_names = ref [] in
    List.iter
      (fun i ->
	already_set_input_names :=
	  set_input
	    ~out_channel ~shot ~seen: !already_set_input_names ~input: i)
      scenario.Basics.test_inputs ;
    (* Appel du composant. *)
    Printf.fprintf out_channel "\n%s\n" Messages.comment_component_call ;
    Printf.fprintf out_channel "%s (" shot.Basics.tested_function ;
    dump_call_args ~out_channel ~args: shot.Basics.tested_parameters ;
    Printf.fprintf out_channel ") ;\n\n" ;
    (* Affichage des rsultats. *)
    Printf.fprintf out_channel "%s\n" Messages.comment_Results_display ;
    dump_scenario_IO_summary ~out_channel ~scenario ;
    (* Vrification des sorties. *)
    Printf.fprintf out_channel "\n%s\n" Messages.comment_Check_outputs ;
    dump_outputs_check ~out_channel ~scenario in
  (* On dumpe enfin tous les tests. *)
  List.iter dump_test shot.Basics.tested_tests
;;





let dump_c_file ~data: shot ~out_channel =
  (* En-tte principale. *)
  Printf.fprintf out_channel "%s\n" Messages.msg_main_header ;
  (* Dclaration des variables globales. *)
  Printf.fprintf out_channel "%s\n" Messages.comment_globvars_declaration ;
  if shot.Basics.tested_globvars = [] then
    Printf.fprintf out_channel "%s\n" Messages.comment_Non_relevant
  else
    List.iter
      (fun vd ->
	dump_variable_declaration
	  ~out_channel ~trailer: " ;\n" ~check_mutability: false
	  ~declaration: vd ~apply_extern: true)
      shot.Basics.tested_globvars ;
  Printf.fprintf out_channel "\n" ;
  (* Gnration du prototype de la fonction teste. *)
  Printf.fprintf out_channel "%s\n" Messages.comment_Component_declaration ;
  Printf.fprintf out_channel "extern void %s (" shot.Basics.tested_function ;
  (* Gnration du prototype de ses arguments. *)
  if shot.Basics.tested_parameters = [] then
    Printf.fprintf out_channel "void"
  else
    dump_parameter_declarations
      ~out_channel ~trailer: ", "
      ~declarations: shot.Basics.tested_parameters ;
  Printf.fprintf out_channel ") ;\n\n" ;
  (* Gnration du main. *)
  Printf.fprintf out_channel "int main ()\n{
int nb_tests = 0;
int nb_tests_OK = 0;
int nb_tests_KO = 0;\n\n" ;
  (* Dclaration des des variables servant  contenir *)
  (* les valeurs  passer comme paramtres.           *)
  Printf.fprintf out_channel "%s\n" Messages.comment_Parameters_declaration ;
  List.iter
    (fun vd ->
      dump_variable_declaration
	~out_channel ~trailer: " ;\n"
	~check_mutability: false ~declaration: vd ~apply_extern: false)
    shot.Basics.tested_parameters ;
  (* Header du rsultat. *)
  Printf.fprintf out_channel "\n%s\n" Messages.comment_Result_header ;
  (* Maintenant, on gnre le code de test proprement dit. *)
  Printf.fprintf
    out_channel "%s\n"
    (Messages.msg_Result_header shot.Basics.tested_function) ;
  dump_tests ~out_channel ~shot ;

  Printf.fprintf out_channel "\n
  printf(\"Synthse des tests.\\n\");
  printf(\"Nombre de tests excuts : %%d\\n\", nb_tests);
  printf(\"Nombre de tests OK : %%d \\n\", nb_tests_OK);
  printf(\"Nombre de tests KO : %%d \\n\", nb_tests_KO);   
  return (0) ;\n}\n"
;;

